#include <math.h>

#include "dds.h"

/****************************************************************************************************/

#define M_PI_F						3.14159265358979323846f

#define DDS_TABLE_SIZE				256	// 8 bits

/****************************************************************************************************/

// sine lookup table
static volatile float dds_sine_table[DDS_TABLE_SIZE];

static volatile enum dds_t_mode dds_mode = SINE_PHASE_SHIFT;

// channel 1
static volatile uint32_t	dds_phase_acc1 = 0;
static volatile uint32_t	dds_tuning_word1;
static volatile float		dds_amplitude1;
static volatile uint8_t		dds_output_enable1 = 0;

// ch1 - sweep parameters
static volatile enum dds_t_sweep_type	dds_sweep_type;
static volatile uint32_t				dds_sweep_init_tuning_word,
										dds_sweep_pulse_samples;
static volatile int32_t					dds_sweep_lin_increment;
static volatile float					dds_sweep_log_param;

// ch1 - burst parameters
static volatile uint32_t	dds_burst_tuning_word;
static volatile float		dds_burst_amplitude1,
							dds_burst_amplitude2;
static volatile uint32_t	dds_burst_period1;

// ch1 - sweep and burst timing
static volatile uint32_t	dds_period_sample,
							dds_period_samples;

// channel 2
static volatile uint32_t	dds_phase_acc2 = 0;
static volatile uint32_t	dds_tuning_word2;
static volatile uint32_t	dds_phase_word2;
static volatile float		dds_amplitude2;
static volatile uint8_t		dds_output_enable2 = 0;

/****************************************************************************************************/

void dds_init(void) {
	// generate sine look up table
	for (uint32_t i = 0; i < DDS_TABLE_SIZE; i++) {
		dds_sine_table[i] = sinf(2.0f * M_PI_F * i / DDS_TABLE_SIZE);
	}
}

/****************************************************************************************************/

static int32_t dds_sine(uint32_t phase, float amplitude) {
	uint16_t index = phase >> 24;	// upper 8 bits of phase

	float fract = (2.0f * M_PI_F * (phase & 0x00ffffff) / 0x1000000) / DDS_TABLE_SIZE;

	// sine = int_sine * fract_cos + int_cos * fract_sine
	float sine = dds_sine_table[index] * (1.0f - 0.5f * fract * fract) +
		dds_sine_table[(index + 64) & 0xff] * fract;

	return (int32_t)roundf(amplitude * sine);
}

/****************************************************************************************************/

void dds_sample(volatile int32_t *channel1, volatile int32_t *channel2) {
	switch (dds_mode) {
	case SINE_PHASE_SHIFT:
		// channel 1
		dds_phase_acc1 += dds_tuning_word1;
		*channel1 = dds_sine(dds_phase_acc1, dds_amplitude1);

		// channel 2
		*channel2 = dds_sine(dds_phase_acc1 + dds_phase_word2, dds_amplitude2);
		break;

	case SINE_SINE:
		// channel 1
		dds_phase_acc1 += dds_tuning_word1;
		*channel1 = dds_sine(dds_phase_acc1, dds_amplitude1);

		// channel 2
		dds_phase_acc2 += dds_tuning_word2;
		*channel2 = dds_sine(dds_phase_acc2, dds_amplitude2);
		break;

	case SWEEP:
		if (dds_sweep_type == LIN) {
			dds_phase_acc1 += dds_sweep_init_tuning_word +
				dds_period_sample * dds_sweep_lin_increment;
		} else {
			dds_phase_acc1 += (uint32_t)roundf(dds_sweep_init_tuning_word
				* powf(dds_sweep_log_param, (float)dds_period_sample / dds_period_samples));
		}

		if (dds_period_sample++ == dds_period_samples) {
			dds_period_sample = 0;
			dds_phase_acc1 = 0;
		}

		*channel1 = dds_sine(dds_phase_acc1, dds_amplitude1);

		// output trigger pulse on channel 2
		if (dds_period_sample >= 0 && dds_period_sample <= dds_sweep_pulse_samples) {
			*channel2 = INT32_MAX;
		} else {
			*channel2 = 0;
		}
		break;

	case BURST:
		if (dds_period_sample++ == dds_period_samples) {
			dds_period_sample = 0;
			dds_phase_acc1 = 0;	// every burst period starts with zero phase
		}

		if (dds_period_sample <= dds_burst_period1) {
			*channel1 = dds_sine(dds_phase_acc1, dds_burst_amplitude1);
			*channel2 = INT32_MAX;	// channel 2 trigger output
		} else {
			*channel1 = dds_sine(dds_phase_acc1, dds_burst_amplitude2);
			*channel2 = 0;			// channel 2 trigger output
		}

		dds_phase_acc1 += dds_burst_tuning_word;
		break;
	}

	if (!dds_output_enable1) {
		*channel1 = 0;
	}
	if (!dds_output_enable2) {
		*channel2 = 0;
	}
}

/****************************************************************************************************/

static int32_t dds_tuning_word(float frequency) {
	return (int32_t)roundf(UINT32_MAX * (frequency / DDS_SAMPLE_RATE));
}

/****************************************************************************************************/

void dds_set_frequency(uint8_t channel, float frequency) {
	if (frequency < DDS_MIN_FREQUENCY || frequency > DDS_MAX_FREQUENCY) {
		return;
	}

	if (channel == 1) {
		dds_tuning_word1 = dds_tuning_word(frequency);
	} else {
		dds_tuning_word2 = dds_tuning_word(frequency);
	}
}

/****************************************************************************************************/

void dds_set_amplitude(uint8_t channel, float amplitude) {
	if (amplitude < 0.0f || amplitude > DDS_MAX_AMPLITUDE) {
		return;
	}

	if (channel == 1) {
		dds_amplitude1 = (amplitude / DDS_MAX_AMPLITUDE) * INT32_MAX;
	} else {
		dds_amplitude2 = (amplitude / DDS_MAX_AMPLITUDE) * INT32_MAX;
	}
}

/****************************************************************************************************/

void dds_set_ch2_phaseshift(float phaseshift) {
	if (phaseshift < 0.0f || phaseshift > DDS_MAX_PHASESHIFT) {
		return;
	}

	dds_phase_word2 = (uint32_t)roundf(UINT32_MAX * (phaseshift / 360));
}

/****************************************************************************************************/

void dds_set_ch1_sweep(enum dds_t_sweep_type type,
	float frequency1, float frequency2, float time) {

	if (frequency1 < DDS_MIN_FREQUENCY || frequency1 > DDS_MAX_FREQUENCY ||
		frequency2 < DDS_MIN_FREQUENCY || frequency2 > DDS_MAX_FREQUENCY ||
		time < DDS_MIN_SWEEP_TIME || (uint32_t)roundf(time * DDS_SAMPLE_RATE) > UINT32_MAX) {
		return;
	}

	dds_sweep_type = type;
	dds_sweep_init_tuning_word = dds_tuning_word(frequency1);

	dds_period_sample = dds_period_samples = (uint32_t)roundf(time * DDS_SAMPLE_RATE);
	dds_sweep_pulse_samples = (uint32_t)roundf(
		((time >= DDS_TRIGGER_PULSE_LENGTH * 2) ? DDS_TRIGGER_PULSE_LENGTH : time * 0.5f)
		* DDS_SAMPLE_RATE
		);

	if (type == LIN) {
		dds_sweep_lin_increment = (int32_t)roundf(
			dds_tuning_word(frequency2 - frequency1) / (time * DDS_SAMPLE_RATE)
		);
	} else {
		dds_sweep_log_param = frequency2 / frequency1;
	}
}

/****************************************************************************************************/

void dds_set_ch1_burst_timed(float frequency,
	float amplitude1, float t1, float amplitude2, float t2) {

	if (frequency < DDS_MIN_FREQUENCY || frequency > DDS_MAX_FREQUENCY ||
		amplitude1 < 0.0f || amplitude1 > DDS_MAX_AMPLITUDE ||
		amplitude2 < 0.0f || amplitude2 > DDS_MAX_AMPLITUDE ||
		t1 < DDS_MIN_BURST_TIME || t2 < DDS_MIN_BURST_TIME ||
		(uint32_t)roundf((t1 + t2) * DDS_SAMPLE_RATE > UINT32_MAX)) {
		return;
	}

	dds_burst_tuning_word = dds_tuning_word(frequency);
	dds_burst_amplitude1 = (amplitude1 / DDS_MAX_AMPLITUDE) * INT32_MAX;
	dds_burst_amplitude2 = (amplitude2 / DDS_MAX_AMPLITUDE) * INT32_MAX;

	dds_burst_period1 = (uint32_t)roundf(t1 * DDS_SAMPLE_RATE);
	dds_period_sample = dds_period_samples = (uint32_t)roundf((t1 + t2) * DDS_SAMPLE_RATE);
}

/****************************************************************************************************/

void dds_set_ch1_burst_cycled(float frequency,
	float amplitude1, uint32_t t1, float amplitude2, uint32_t t2) {

	if (frequency < DDS_MIN_FREQUENCY || frequency > DDS_MAX_FREQUENCY ||
		amplitude1 < 0.0f || amplitude1 > DDS_MAX_AMPLITUDE ||
		amplitude2 < 0.0f || amplitude2 > DDS_MAX_AMPLITUDE ||
		t1 < 1 || t2 < 1) {
		return;
	}

	uint32_t cycle_samples = (uint32_t)roundf(DDS_SAMPLE_RATE / frequency);
	if ((t1 + t2) * cycle_samples > UINT32_MAX) {
		return;
	}

	dds_burst_tuning_word = dds_tuning_word(frequency);
	dds_burst_amplitude1 = (amplitude1 / DDS_MAX_AMPLITUDE) * INT32_MAX;
	dds_burst_amplitude2 = (amplitude2 / DDS_MAX_AMPLITUDE) * INT32_MAX;

	dds_burst_period1 = t1 * cycle_samples;
	dds_period_sample = dds_period_samples = (t1 + t2) * cycle_samples;
}

/****************************************************************************************************/

void dds_set_mode(enum dds_t_mode mode) {
	dds_mode = mode;
}

/****************************************************************************************************/

void dds_set_output_enable(uint8_t channel, uint8_t output_enable) {
	if (channel == 1) {
		dds_output_enable1 = output_enable;
	} else {
		dds_output_enable2 = output_enable;
	}
}

/****************************************************************************************************/
