#include "stm32f30x.h"
#include "delay.h"
#include "display.h"

/****************************************************************************************************/

#define LCD_INST_CLEAR_DISPLAY			0x01
#define LCD_INST_ENTRY_MODE_SET			0x06	// DDRAM address increment, disable display shift
#define LCD_INST_DISPLAY_ON				0x0c	// display on, cursor off, blinking off
#define LCD_INST_BLINK_ON				0x0d	// display on, cursor off, blinking on
#define LCD_INST_SET_DDRAM_ADDR			0x80

/****************************************************************************************************/

enum disp_t_transfer_type { INST, DATA };

/****************************************************************************************************/

static const uint8_t disp_line_address[] = { 0, 0x40, 0x14, 0x54 };

static volatile int32_t disp_cursor_position = -1;

/****************************************************************************************************/

static inline void display_strobe_nibble(uint8_t nibble) {
	GPIOB->BSRR = GPIO_BSRR_BS_11;		// E=1
	delay_us(1);						// >460ns pulse length
	GPIOA->ODR = (nibble & 0x0f) << 8;	// output data
	delay_us(1);						// >80ns data setup time
	GPIOB->BSRR = GPIO_BSRR_BR_11;		// E=0
}

/****************************************************************************************************/

static void display_wait_until_busy(void) {
	// poll display busy flag
	GPIOA->MODER &= ~(GPIO_MODER_MODER8 | GPIO_MODER_MODER9 |
		GPIO_MODER_MODER10 | GPIO_MODER_MODER11);		// set data pins to input mode
	GPIOB->BSRR = GPIO_BSRR_BS_10 | GPIO_BSRR_BR_2;	// R/W=1, RS=0

	uint32_t busy = 1;
	while (busy) {
		// read first nibble
		GPIOB->BSRR = GPIO_BSRR_BS_11;		// E=1
		delay_us(1);						// >480ns data setup time and pulse length
		busy = GPIOA->IDR & GPIO_IDR_11;	// read busy flag
		GPIOB->BSRR = GPIO_BSRR_BR_11;		// E=0

		// read second nibble
		GPIOB->BSRR = GPIO_BSRR_BS_11;		// E=1
		delay_us(1);						// >480ns pulse length
		GPIOB->BSRR = GPIO_BSRR_BR_11;		// E=0
	}

	GPIOB->BSRR = GPIO_BSRR_BR_10;			// R/W=0

	// set data pins to output mode
	GPIOA->MODER |= GPIO_MODER_MODER8_0 | GPIO_MODER_MODER9_0 |
		GPIO_MODER_MODER10_0 | GPIO_MODER_MODER11_0;
}

/****************************************************************************************************/

static void display_send(enum disp_t_transfer_type type, uint8_t byte) {
	display_wait_until_busy();

	if (type == DATA) {
		GPIOB->BSRR = GPIO_BSRR_BS_2;	// RS=1
		delay_us(1);					// 30ns setup time
	}

	display_strobe_nibble(byte >> 4);	// output high nibble of data byte
	display_strobe_nibble(byte);		// output low nibble of data byte
}

/****************************************************************************************************/

void display_write(const int8_t *str) {
	if (disp_cursor_position != -1) {
		display_send(INST, LCD_INST_DISPLAY_ON);	// hide cursor
	}

	display_send(INST, LCD_INST_CLEAR_DISPLAY);

	uint32_t line = 0;
	while (*str) {
		if (*str == '\n') {
			// change DDRAM addres to the begining of the next line
			display_send(INST, LCD_INST_SET_DDRAM_ADDR + disp_line_address[++line]);
		} else {
			display_send(DATA, *str);
		}
		str++;
	}

	// show cursor again, if was shown
	if (disp_cursor_position != -1) {
		display_send(INST, LCD_INST_SET_DDRAM_ADDR + disp_cursor_position);
		display_send(INST, LCD_INST_BLINK_ON);
	}
}

/****************************************************************************************************/

void display_show_cursor(uint8_t position) {
	disp_cursor_position = disp_line_address[position / 20] + (position % 20);

	// set DDRAM address
	display_send(INST, LCD_INST_SET_DDRAM_ADDR + disp_cursor_position);
	display_send(INST, LCD_INST_BLINK_ON);	// turn on cursor blinking
}

/****************************************************************************************************/

void display_hide_cursor(void) {
	display_send(INST, LCD_INST_DISPLAY_ON);
	disp_cursor_position = -1;
}

/****************************************************************************************************/

void display_init(void) {
	// setup pins PA8-PA11 as medium speed outputs
	GPIOA->MODER |= GPIO_MODER_MODER8_0 | GPIO_MODER_MODER9_0 |
		GPIO_MODER_MODER10_0 | GPIO_MODER_MODER11_0;
	GPIOA->OSPEEDR |= GPIO_OSPEEDER_OSPEEDR8_0 | GPIO_OSPEEDER_OSPEEDR9_0 |
		GPIO_OSPEEDER_OSPEEDR10_0 | GPIO_OSPEEDER_OSPEEDR11_0;

	// setup pins PB2, PB10 and PB11 as medium speed outputs
	GPIOB->MODER |= GPIO_MODER_MODER2_0 | GPIO_MODER_MODER10_0 | GPIO_MODER_MODER11_0;
	GPIOB->OSPEEDR |= GPIO_OSPEEDER_OSPEEDR2_0 | GPIO_OSPEEDER_OSPEEDR10_0 |
		GPIO_OSPEEDER_OSPEEDR11_0;

	/*
	 * perform 4bit communication initialization sequence
	 * (according to ST7066 controller datasheet)
	 */
	delay_us(30000);				// wait 30ms

	display_strobe_nibble(0x03);	// function set 1
	delay_us(8200);

	display_strobe_nibble(0x03);	// function set 2
	delay_us(200);

	display_strobe_nibble(0x03);	// function set 3
	display_wait_until_busy();

	display_strobe_nibble(0x02);
	display_wait_until_busy();

	// function set instruction - 4 bit interface, 2 lines, 5x8 font
	display_strobe_nibble(0x02);
	display_strobe_nibble(0x08);
	display_wait_until_busy();

	display_send(INST, LCD_INST_DISPLAY_ON);		// display on
	display_send(INST, LCD_INST_CLEAR_DISPLAY);		// clear display
	display_send(INST, LCD_INST_ENTRY_MODE_SET);	// entry mode set
}

/****************************************************************************************************/
