#include "stm32f30x.h"
#include "i2c.h"

/****************************************************************************************************/

#define I2C_EEPROM_ADDRESS			0xa0
#define I2C_CS4398_ADDRESS			0x98

/****************************************************************************************************/

static volatile uint16_t	i2c_address;
static volatile uint8_t		i2c_transfer_size,
							i2c_wait;

/****************************************************************************************************/

void i2c_event_handler(void) {
	if (I2C1->ISR & I2C_ISR_NACKF) {
		// NACK was received
		I2C1->ICR = I2C_ICR_NACKCF | I2C_ICR_STOPCF;	// clear event flag

		while (I2C1->ISR & I2C_ISR_BUSY);

		I2C1->CR2 |= I2C_CR2_START;	// start transfer again (while polling memory)
		return;
	}

	if (I2C1->ISR & I2C_ISR_TXIS) {
		I2C1->TXDR = (uint8_t)i2c_address;	// send address byte
		I2C1->CR1 &= ~I2C_CR1_TXIE;			// disable transmit event

		I2C1->ICR = I2C_ICR_STOPCF;			// clear STOP event flag
		I2C1->CR1 |= I2C_CR1_STOPIE;		// enable STOP event
		return;
	}

	if (I2C1->ISR & I2C_ISR_STOPF) {
		// STOP condition detected
		// disable DMA requests and STOP condition event
		I2C1->CR1 &= ~(I2C_CR1_RXDMAEN | I2C_CR1_TXDMAEN | I2C_CR1_STOPIE);
		I2C1->CR1 |= I2C_CR1_TXIE;	// enable transmit event
		I2C1->ICR = I2C_ICR_STOPCF;	// clear interrupt flag

		i2c_wait = 0;				// release waiting thread
		return;
	}

	if (I2C1->ISR & I2C_ISR_TCR) {
		// transfer complete reload
		I2C1->CR1 |= I2C_CR1_TXDMAEN;								// enable DMA request
		I2C1->CR2 = (i2c_transfer_size << 16) | I2C_CR2_AUTOEND;	// set transfer size
		return;
	}

	if (I2C1->ISR & I2C_ISR_TC) {
		// transfer complete
		I2C1->CR1 |= I2C_CR1_RXDMAEN;				// enable DMA request
		I2C1->CR2 = ((i2c_address >> 7) & 0xff) |	// setup slave address
			(i2c_transfer_size << 16) |				// setup transfer length
			I2C_CR2_AUTOEND |						// automatically generate STOP condition
			I2C_CR2_RD_WRN |						// read direction
			I2C_CR2_START;
	}
}

/****************************************************************************************************/

void i2c_init(void) {
	RCC->APB1ENR |= RCC_APB1ENR_I2C1EN;	// enable I2C1 periphery clock

	// setup pin PB8 as I2C1_SCL and PB9 as I2C1_SDA
	GPIOB->AFR[1] |= 0x44;
	GPIOB->MODER |= GPIO_MODER_MODER8_1 | GPIO_MODER_MODER9_1;

	I2C1->TIMINGR = 0x10c08dcf;	// setup timing - 100kHz Standard-mode

	// enable TC event, STOP event, NACK event, TX event and enable I2C periphery
	I2C1->CR1 = I2C_CR1_TCIE | I2C_CR1_NACKIE | I2C_CR1_TXIE | I2C_CR1_PE;

	NVIC->IP[31] = 0x10;
	NVIC_EnableIRQ(31);	// enable I2C1 event interrupt

	RCC->AHBENR |= RCC_AHBENR_DMA1EN;	// enable DMA1 periphery clock
}

/****************************************************************************************************/

static void i2c_read(uint8_t slave_addr, uint16_t addr, uint8_t *buffer, uint8_t size) {
	// sequential read of maximum of 255 bytes in address range with same upper 3 bits

	i2c_address = ((uint16_t)slave_addr << 7) | addr;
	i2c_transfer_size = size;

	// setup DMA transfer
	DMA1_Channel7->CPAR = (uint32_t)&(I2C1->RXDR);
	DMA1_Channel7->CMAR = (uint32_t)buffer;
	DMA1_Channel7->CNDTR = i2c_transfer_size;
	// 8 bits from peripheral to memory, memory increment mode, very high priority
	DMA1_Channel7->CCR = DMA_CCR_PL | DMA_CCR_MINC | DMA_CCR_EN;

	i2c_wait = 1;

	I2C1->CR2 = ((i2c_address >> 7) & 0xff) |	// setup slave address
		(1 << 16) |								// setup transfer length
		I2C_CR2_START;							// start transfer

	while (i2c_wait);

	DMA1_Channel7->CCR = 0;	// disable DMA
}

/****************************************************************************************************/

static void i2c_write(uint8_t slave_addr, uint16_t addr, const uint8_t *data, uint8_t size) {
	i2c_address = ((uint16_t)slave_addr << 7) | addr;
	i2c_transfer_size = size;

	// setup DMA transfer
	DMA1_Channel6->CPAR = (uint32_t)&(I2C1->TXDR);
	DMA1_Channel6->CMAR = (uint32_t)data;
	DMA1_Channel6->CNDTR = i2c_transfer_size;
	// 8 bits from memory to peripheral, memory increment mode, very high priority
	DMA1_Channel6->CCR = DMA_CCR_PL | DMA_CCR_MINC | DMA_CCR_DIR | DMA_CCR_EN;

	i2c_wait = 1;

	I2C1->CR2 = ((i2c_address >> 7) & 0xff) |				// setup slave address
		(1 << 16) |											// setup transfer length
		I2C_CR2_RELOAD | I2C_CR2_AUTOEND | I2C_CR2_START;

	while (i2c_wait);	// wait until transfer is done

	DMA1_Channel6->CCR = 0;	// disable DMA
}

/****************************************************************************************************/

void memory_write(uint16_t addr, const uint8_t *data, uint8_t size) {
	while (size > 0) {
		i2c_write(I2C_EEPROM_ADDRESS, addr, data, (size > 16) ? 16 : size);

		addr += 16;							// increment address
		data += 16;							// increment data pointer
		size = (size > 16) ? size - 16 : 0;	// decrement data counter
	}
}

/****************************************************************************************************/

void memory_read(uint16_t addr, uint8_t *buffer, uint8_t size) {
	i2c_read(I2C_EEPROM_ADDRESS, addr, buffer, size);
}

/****************************************************************************************************/

void dac_write_register(uint8_t addr, uint8_t data) {
	i2c_write(I2C_CS4398_ADDRESS, addr, &data, 1);
}

/****************************************************************************************************/

uint8_t dac_read_register(uint8_t addr) {
	uint8_t data;

	i2c_read(I2C_CS4398_ADDRESS, addr, &data, 1);
	return data;
}

/****************************************************************************************************/
