#include <stdint.h>

#include "stm32f30x.h"
#include "user_interface.h"
#include "delay.h"
#include "keyboard.h"

/****************************************************************************************************/

#define BUTTON_ROWS						5
#define BUTTON_COLS						6

#define BUTTON_SCAN_TIME				5		// ms
#define BUTTON_DEBOUNCE_PRESS			10		// ms
#define BUTTON_DEBOUNCE_RELEASE			50		// ms

/****************************************************************************************************/

struct kbd_t_button {
	uint8_t	state,
			debounce_timer;
};

/****************************************************************************************************/

static volatile struct kbd_t_button buttons[BUTTON_ROWS][BUTTON_COLS];

/****************************************************************************************************/

void kbd_scan_handler(void) {
	uint16_t button_state;

	// scan keyboard
	for (uint32_t row = 0; row < BUTTON_ROWS; row++) {
		GPIOB->ODR &= ~(0x08 << row);	// power up row

		for (uint32_t col = 0; col < BUTTON_COLS; col++) {
			button_state = !(GPIOA->IDR & (0x04 << col));
			if (button_state == buttons[row][col].state) {
				// button state is unchanged
				buttons[row][col].debounce_timer = (button_state) ?
					BUTTON_DEBOUNCE_RELEASE / BUTTON_SCAN_TIME :
					BUTTON_DEBOUNCE_PRESS / BUTTON_SCAN_TIME;
			} else {
				// button state has changed
				if (--buttons[row][col].debounce_timer == 0) {
					// debouncing complete
					buttons[row][col].state = button_state;
					if (button_state) {
						gen_process_button(8 * row + col);
					}

					buttons[row][col].debounce_timer = (button_state) ?
						BUTTON_DEBOUNCE_RELEASE / BUTTON_SCAN_TIME :
						BUTTON_DEBOUNCE_PRESS / BUTTON_SCAN_TIME;
				}
			}
		}

		GPIOB->ODR |= 0xf8;	// power down row
	}

	TIM6->SR = 0;	// clear interrupt flag
}

/****************************************************************************************************/

void tim2_event_handler(void) {
	if (TIM2->CR1 & TIM_CR1_DIR) {
		gen_process_button(BUTTON_ROTATE_LEFT);
	} else {
		gen_process_button(BUTTON_ROTATE_RIGHT);
	}

	TIM2->SR &= ~TIM_SR_UIF;	// clear update interrupt flag
}

/****************************************************************************************************/

void kbd_init(void) {
	// pins PA2-PA7 are defaultly setup as inputs

	// setup pins PB3-PB7 as medium speed open drain outputs
	GPIOB->BSRR = GPIO_BSRR_BS_3 | GPIO_BSRR_BS_4 | GPIO_BSRR_BS_5 |
			GPIO_BSRR_BS_6 | GPIO_BSRR_BS_7;
	GPIOB->OTYPER |= GPIO_OTYPER_OT_3 | GPIO_OTYPER_OT_4 | GPIO_OTYPER_OT_5 |
			GPIO_OTYPER_OT_6 | GPIO_OTYPER_OT_7;
	GPIOB->OSPEEDR &= ~GPIO_OSPEEDER_OSPEEDR3;
	GPIOB->OSPEEDR |= GPIO_OSPEEDER_OSPEEDR3_0 | GPIO_OSPEEDER_OSPEEDR4_0 | GPIO_OSPEEDER_OSPEEDR5_0 |
			GPIO_OSPEEDER_OSPEEDR6_0 | GPIO_OSPEEDER_OSPEEDR7_0;

	GPIOB->MODER &= 0xffff003f;
	GPIOB->MODER |= GPIO_MODER_MODER3_0 | GPIO_MODER_MODER4_0 | GPIO_MODER_MODER5_0 |
		GPIO_MODER_MODER6_0 | GPIO_MODER_MODER7_0;

	// initialize buttons array
	for (uint32_t row = 0; row < BUTTON_ROWS; row++) {
		for (uint32_t col = 0; col < BUTTON_COLS; col++) {
			buttons[row][col].state = 0;
			buttons[row][col].debounce_timer = BUTTON_DEBOUNCE_PRESS / BUTTON_SCAN_TIME;
		}
	}

	// initialize scanning timer
	RCC->APB1ENR |= RCC_APB1ENR_TIM6EN;

	TIM6->PSC	= 7199;	// timer clock = 10kHz
	TIM6->ARR	= 10 * BUTTON_SCAN_TIME;
	TIM6->DIER	= TIM_DIER_UIE;
	TIM6->CR1	= TIM_CR1_CEN;

	NVIC->IP[54] = 0x20;	// set timer interrupt priority
	NVIC_EnableIRQ(54);		// enable timer interrupt

	// initialize rotary encoder
	RCC->APB1ENR |= RCC_APB1ENR_TIM2EN;

	// setup pins PA0 and PA1 as TIM2 timer inputs
	GPIOA->AFR[0] |= 0x11;
	GPIOA->MODER |= GPIO_MODER_MODER0_1 | GPIO_MODER_MODER1_1;

	TIM2->CCMR1 = TIM_CCMR1_IC1F | TIM_CCMR1_CC1S_0;
	TIM2->SMCR = TIM_SMCR_SMS_0 | TIM_SMCR_SMS_1;

	TIM2->ARR = 2;
	TIM2->DIER = TIM_DIER_UIE;

	NVIC->IP[28] = 0x20;
	NVIC_EnableIRQ(28);

	TIM2->CR1 = TIM_CR1_CEN;
}

/****************************************************************************************************/
