/*****************************************************************************************************

	AUDIO SIGNAL GENERATOR

	Lukas Kroulik - January-May/2017

	to be run on STM32F301C8T with 8MHz HSE crystal

 ****************************************************************************************************

	PIN ASSIGMENT:

	PA12	master clock input
	PB15	I2S data
	PB13	I2S clock
	PB12	I2S word select

	PB14	DAC reset

	PA8		display DB4
	PA9		display DB5
	PA10	display DB6
	PA11	display DB7
	PB2		display RS
	PB10	display	R/W
	PB11	display E

	PA2		keyboard COL0
	PA3		keyboard COL1
	PA4		keyboard COL2
	PA5		keyboard COL3
	PA6		keyboard COL4
	PA7		keyboard COL5

	PB3		keyboard ROW0
	PB4		keyboard ROW1
	PB5		keyboard ROW2
	PB6		keyboard ROW3
	PB7		keyboard ROW4

	PA0		rotary encoder 1
	PA1		rotary encoder 2

	PB0		LED 1
	PB1		LED 2

	PB8		I2C SCL
	PB9		I2C SDA

	PA13	SWDIO
	PA14	SWDCLK

	PC13	unused (setup as input with pulldown)
	PC14	unused (setup as input with pulldown)
	PC15	unused (setup as input - pin is grounded on pcb)
	PA15	unused (setup as input with pullup)

*****************************************************************************************************/

#include "stm32f30x.h"
#include "dds.h"
#include "delay.h"
#include "display.h"
#include "keyboard.h"
#include "user_interface.h"
#include "leds.h"
#include "i2c.h"

/****************************************************************************************************/

#define BUFFER_SIZE			(2 * DDS_SAMPLE_RATE / 50)	// 20ms total buffer length

/****************************************************************************************************/

static volatile int32_t buffer[BUFFER_SIZE];

/****************************************************************************************************/

void dma_ch5_event_handler(void) {
	if (DMA1->ISR &= DMA_ISR_HTIF5) {
		// half transfer - fill top half of buffer
		for (uint32_t i = 0; i < BUFFER_SIZE / 2; i += 2) {
			dds_sample(&buffer[i+1], &buffer[i]);
			buffer[i] = __REV16(__REV(buffer[i]));
			buffer[i+1] = __REV16(__REV(buffer[i+1]));
		}
	} else {
		// transfer complete - fill buttom half of buffer
		for (uint32_t i = BUFFER_SIZE / 2; i < BUFFER_SIZE; i += 2) {
			dds_sample(&buffer[i+1], &buffer[i]);
			buffer[i] = __REV16(__REV(buffer[i]));
			buffer[i+1] = __REV16(__REV(buffer[i+1]));
		}
	}

	DMA1->IFCR = DMA_IFCR_CGIF5;	// clear all interrupt flags
}

/****************************************************************************************************/

void i2s_init(void) {
	/*
	 * setup pin PA12 as I2SCKIN (master clock input), PB12 as I2S2_WS,
	 * PB13 as I2S2_CK and PB15 as I2S2_SD
	 */
	GPIOA->AFR[1] |= 0x50000;
	GPIOA->MODER |= GPIO_MODER_MODER12_1;

	GPIOB->AFR[1] |= 0x50550000;
	GPIOB->MODER |= GPIO_MODER_MODER12_1 | GPIO_MODER_MODER13_1 | GPIO_MODER_MODER15_1;

	RCC->CFGR |= RCC_CFGR_I2SSRC;		// setup I2SCKIN as clock for I2S

	RCC->APB1ENR |= RCC_APB1ENR_SPI2EN;	// enable I2S periphery clock
	RCC->AHBENR |= RCC_AHBENR_DMA1EN;	// enable DMA1 periphery clock

	// initialize I2S
	SPI2->I2SPR = 4;	// bit clock = I2SCKIN / 8

	/*
	 * master - transmit mode, MSB justified standard, inactive clock polarity high,
	 * 24 bit data length, 32 bit channel length
	 */
	SPI2->I2SCFGR = SPI_I2SCFGR_I2SMOD | SPI_I2SCFGR_I2SCFG_1 | SPI_I2SCFGR_I2SSTD_0 |
		SPI_I2SCFGR_CKPOL | SPI_I2SCFGR_DATLEN_0;
	SPI2->CR2 |= SPI_CR2_TXDMAEN;	// enable DMA request

	DMA1_Channel5->CPAR = (uint32_t)&(SPI2->DR);
	DMA1_Channel5->CMAR = (uint32_t)buffer;
	DMA1_Channel5->CNDTR = 2 * BUFFER_SIZE;

	/*
	 * 16 bits from memory to peripheral, memory increment mode, circular mode,
	 * half transfer and transfer complete interrupts, very high priority
	 */
	DMA1_Channel5->CCR = DMA_CCR_PL | DMA_CCR_MSIZE_0 | DMA_CCR_PSIZE_0 |
		DMA_CCR_MINC | DMA_CCR_CIRC | DMA_CCR_DIR | DMA_CCR_HTIE |
		DMA_CCR_TCIE | DMA_CCR_EN;

	NVIC_EnableIRQ(15);	// enable DMA interrupt

	SPI2->I2SCFGR |= SPI_I2SCFGR_I2SE;	// enable I2S
}

/****************************************************************************************************/

void misc_init(void) {
	// enable GPIOA-GPIOC periphery clocks
	RCC->AHBENR |= RCC_AHBENR_GPIOAEN | RCC_AHBENR_GPIOBEN | RCC_AHBENR_GPIOCEN;

	// configure unused pins
	GPIOB->PUPDR = 0;
	GPIOB->MODER = 0;

	// activate pulldowns on unused pins PC13 and PC14
	GPIOC->PUPDR |= GPIO_PUPDR_PUPDR13_1 | GPIO_PUPDR_PUPDR14_1;

	// setup pin PA15 as input with pullup
	GPIOA->MODER &= ~GPIO_MODER_MODER15;

	// setup SysTick timer interrupt priority - used by delay_us function
	SCB->SHP[11] = 0x10;
}

/****************************************************************************************************/

void dac_init(void) {
	GPIOB->MODER |= GPIO_MODER_MODER14_0;	// setup pin PB14 as output
	GPIOB->BSRR = GPIO_BSRR_BS_14;			// disable DAC reset pin

	delay_us(1);

	// enable DAC control port
	dac_write_register(0x08, 0xc0);	// Misc. control register, PDN=1, CPEN=1
	dac_write_register(0x03, 0x29);	// invert channel B signal polarity
	dac_write_register(0x04, 0x00);	// disable Auto-Mute
	dac_write_register(0x08, 0x40);	// enable DAC
}

/****************************************************************************************************/

int main(void) {

	misc_init();

	i2c_init();

	display_init();
	leds_init();
	kbd_init();

	gen_init();

	// fill whole buffer with samples
	for (uint32_t i = 0; i < BUFFER_SIZE; i += 2) {
		dds_sample(&buffer[i+1], &buffer[i]);
	}

	dac_init();

	i2s_init();

	while (1);

	return 0;
}

/****************************************************************************************************/
