#include <stdio.h>
#include <math.h>

#include "dds.h"
#include "display.h"
#include "leds.h"
#include "i2c.h"

#include "user_interface.h"

/****************************************************************************************************/

#define M_SQRT2_F				1.41421356237309504880f

// buttons
#define BUTTON_MODE				0
#define BUTTON_AMP_UNIT			34
#define BUTTON_STORE_RECALL		37

#define BUTTON_F1				8
#define BUTTON_F2				16
#define BUTTON_F3				24
#define BUTTON_F4				1
#define BUTTON_F5				9
#define BUTTON_F6				17
#define BUTTON_F7				25

#define BUTTON_OUTPUT_ENABLE_1	35
#define BUTTON_OUTPUT_ENABLE_2	36

#define BUTTON_LEFT				32
#define BUTTON_RIGHT			33

#define BUTTON_DECIMAL_POINT	26
#define BUTTON_ENTER			28

#define BUTTON_0				27
#define BUTTON_1				2
#define BUTTON_2				3
#define BUTTON_3				4
#define BUTTON_4				10
#define BUTTON_5				11
#define BUTTON_6				12
#define BUTTON_7				18
#define BUTTON_8				19
#define BUTTON_9				20

// function modes
#define FUNCTION_GENERATOR		0
#define FUNCTION_SETTING_RECALL	1
#define FUNCTION_SETTING_STORE	2

// generator modes
#define MODE_SINE_PHASE_SHIFT	0
#define MODE_SINE_SINE			1
#define MODE_SWEEP				2
#define MODE_BURST				3

// parameters
#define MODE1_FREQUENCY			0
#define MODE1_PHASESHIFT		1
#define MODE1_AMPLITUDE1		2
#define MODE1_AMPLITUDE2		3

#define MODE2_FREQUENCY1		4
#define MODE2_FREQUENCY2		5
#define MODE2_AMPLITUDE1		6
#define MODE2_AMPLITUDE2		7

#define MODE3_FREQUENCY1		8
#define MODE3_FREQUENCY2		9
#define MODE3_TIME				10
#define MODE3_AMPLITUDE			11

#define MODE4_FREQUENCY			12
#define MODE4_AMPLITUDE1		13
#define MODE4_CYCLES1			14
#define MODE4_TIME1				15
#define MODE4_AMPLITUDE2		16
#define MODE4_CYCLES2			17
#define MODE4_TIME2				18

/****************************************************************************************************/

enum gen_t_amp_unit { VRMS, VPEAK, DBFS };

enum gen_t_burst_mode { TIME, CYCLES };

enum gen_t_input_mode { ENCODER, NUMERIC };

enum gen_t_input_part { INT, FRACT };

struct gen_t_param {
	float	value;
	float	min_value,
			max_value;
	uint8_t	int_size,
			fract_size;
	uint8_t	display_pos;
};

/****************************************************************************************************/

static volatile uint8_t gen_function_mode	= FUNCTION_GENERATOR;
static volatile uint8_t gen_mode			= MODE_SINE_PHASE_SHIFT;

static volatile enum gen_t_amp_unit		gen_amp_unit	= VRMS;
static volatile enum dds_t_sweep_type	gen_sweep_type	= LOG;
static volatile enum gen_t_burst_mode	gen_burst_mode	= CYCLES;

static volatile uint8_t	gen_output_enable1 = 0,
						gen_output_enable2 = 0;

static volatile struct gen_t_param gen_params[] = {
	// mode 1 - sine/phase shifted sine parameters
	// MODE1_FREQUENCY
	{ 937.5f, DDS_MIN_FREQUENCY, DDS_MAX_FREQUENCY, 5, 3, 12 },
	// MODE1_PHASESHIFT
	{ 0.0f, 0.0f, DDS_MAX_PHASESHIFT, 3, 3, 52 },
	// MODE1_AMPLITUDE1
	{ DDS_MAX_AMPLITUDE, 0.0f, DDS_MAX_AMPLITUDE, 1, 2, 32 },
	// MODE1_AMPLITUDE2
	{ DDS_MAX_AMPLITUDE, 0.0f, DDS_MAX_AMPLITUDE, 1, 2, 72 },

	// mode 2 - sine/sine parameters
	// MODE2_FREQUENCY1
	{ 1000.0f, DDS_MIN_FREQUENCY, DDS_MAX_FREQUENCY, 5, 3, 12 },
	// MODE2_FREQUENCY2
	{ 2000.0f, DDS_MIN_FREQUENCY, DDS_MAX_FREQUENCY, 5, 3, 52 },
	// MODE2_AMPLITUDE1
	{ DDS_MAX_AMPLITUDE, 0.0f, DDS_MAX_AMPLITUDE, 1, 2, 32 },
	// MODE2_AMPLITUDE2
	{ DDS_MAX_AMPLITUDE, 0.0f, DDS_MAX_AMPLITUDE, 1, 2, 72 },

	// mode 3 - sweep parameters
	// MODE3_FREQUENCY1
	{ 100.0f, DDS_MIN_FREQUENCY, DDS_MAX_FREQUENCY, 5, 3, 12 },
	// MODE3_FREQUENCY2
	{ 10000.0f, DDS_MIN_FREQUENCY, DDS_MAX_FREQUENCY, 5, 3, 32 },
	// MODE3_TIME
	{ .5f, DDS_MIN_SWEEP_TIME, 9.999f, 1, 3, 40 },
	// MODE3_AMPLITUDE
	{ DDS_MAX_AMPLITUDE, 0.0f, DDS_MAX_AMPLITUDE, 1, 2, 52 },

	// mode 4 - burst parameters
	// MODE4_FREQUENCY
	{ 1000.0f, DDS_MIN_FREQUENCY, DDS_MAX_FREQUENCY, 5, 3, 13 },
	// MODE4_AMPLITUDE1
	{ DDS_MAX_AMPLITUDE, 0.0f, DDS_MAX_AMPLITUDE, 1, 2, 22 },
	// MODE4_CYCLES1
	{ 10.0f, 1.0f, 99999.0f, 5, 0, 36  },
	// MODE4_TIME1
	{ 0.01f, DDS_MIN_BURST_TIME, 9.999f, 1, 3, 33 },
	// MODE4_AMPLITUDE2
	{ 0.0f, 0.0f, DDS_MAX_AMPLITUDE, 1, 2, 42 },
	// MODE4_CYCLES2
	{ 10.0f, 1.0f, 99999.0f, 5, 0, 56 },
	// MODE4_TIME2
	{ 0.01f, DDS_MIN_BURST_TIME, 9.999f, 1, 3, 53 }
};

static volatile enum gen_t_input_mode	gen_input_mode;
static volatile struct gen_t_param		*gen_input_param = NULL;	// parameter being currently input
static volatile float					gen_input_param_prev;

static volatile enum gen_t_input_part	gen_input_part;
static volatile uint8_t					gen_input_pos;

static volatile int8_t					gen_cursor_exponent;

/****************************************************************************************************/

// declarations for setting storage functions
#define GEN_PARAMS_COUNT		sizeof gen_params / sizeof (struct gen_t_param)

struct gen_t_setting {
	uint8_t					gen_mode;
	enum gen_t_amp_unit		gen_amp_unit;
	enum dds_t_sweep_type	gen_sweep_type;
	enum gen_t_burst_mode	gen_burst_mode;
	float					gen_param_values[GEN_PARAMS_COUNT];
};

/****************************************************************************************************/

static float gen_scale_amplitude(float amp) {
	switch (gen_amp_unit) {
	case VRMS:	return amp;
	case VPEAK:	return ( amp / M_SQRT2_F );
	default:	return ( DDS_MAX_AMPLITUDE * powf(10.0f, amp / 20.0f) );
	}
}

/****************************************************************************************************/

static void gen_set_params(void) {
	switch (gen_mode) {
	case MODE_SINE_PHASE_SHIFT:
		dds_set_frequency(1, gen_params[MODE1_FREQUENCY].value);
		dds_set_ch2_phaseshift(gen_params[MODE1_PHASESHIFT].value);
		dds_set_amplitude(1, gen_scale_amplitude(gen_params[MODE1_AMPLITUDE1].value));
		dds_set_amplitude(2, gen_scale_amplitude(gen_params[MODE1_AMPLITUDE2].value));
		break;

	case MODE_SINE_SINE:
		dds_set_frequency(1, gen_params[MODE2_FREQUENCY1].value);
		dds_set_frequency(2, gen_params[MODE2_FREQUENCY2].value);
		dds_set_amplitude(1, gen_scale_amplitude(gen_params[MODE2_AMPLITUDE1].value));
		dds_set_amplitude(2, gen_scale_amplitude(gen_params[MODE2_AMPLITUDE2].value));
		break;

	case MODE_SWEEP:
		dds_set_ch1_sweep(gen_sweep_type, gen_params[MODE3_FREQUENCY1].value,
			gen_params[MODE3_FREQUENCY2].value, gen_params[MODE3_TIME].value
		);
		dds_set_amplitude(1, gen_scale_amplitude(gen_params[MODE3_AMPLITUDE].value));
		break;

	case MODE_BURST:
		if (gen_burst_mode == TIME) {
			dds_set_ch1_burst_timed(gen_params[MODE4_FREQUENCY].value,
				gen_scale_amplitude(gen_params[MODE4_AMPLITUDE1].value),
				gen_params[MODE4_TIME1].value,
				gen_scale_amplitude(gen_params[MODE4_AMPLITUDE2].value),
				gen_params[MODE4_TIME2].value
			);
		} else {
			dds_set_ch1_burst_cycled(gen_params[MODE4_FREQUENCY].value,
				gen_scale_amplitude(gen_params[MODE4_AMPLITUDE1].value),
				(uint32_t)gen_params[MODE4_CYCLES1].value,
				gen_scale_amplitude(gen_params[MODE4_AMPLITUDE2].value),
				(uint32_t)gen_params[MODE4_CYCLES2].value
			);
		}
		break;

	}
}

/****************************************************************************************************/

static void gen_update_display(void) {
	if (gen_function_mode == FUNCTION_SETTING_RECALL) {
		display_write((int8_t*)"SETTING RECALL\n\xc7select memory slot\n 0-9");
		return;
	} else {
		if (gen_function_mode == FUNCTION_SETTING_STORE) {
			display_write((int8_t*)"SETTING STORE\n\xc7select memory slot\n 0-9");
			return;
		}
	}

	const int8_t *amp_unit;
	switch (gen_amp_unit) {
	case VRMS:	amp_unit = (const int8_t*)"Vrms";	break;
	case VPEAK:	amp_unit = (const int8_t*)"Vp";		break;
	default:	amp_unit = (const int8_t*)"dBFS";	break;
	}

	int8_t buffer[84];
	switch (gen_mode) {
	case MODE_SINE_PHASE_SHIFT:
		sprintf((char*)buffer,
			"1 SINE  %#9.3fHz\n"
			"          %#6.2f%-4s\n"
			"2 PHASE   %#7.3f\xb2\n"
			"          %#6.2f%-4s",
			gen_params[MODE1_FREQUENCY].value, gen_params[MODE1_AMPLITUDE1].value, amp_unit,
			gen_params[MODE1_PHASESHIFT].value, gen_params[MODE1_AMPLITUDE2].value, amp_unit
		);
		break;

	case MODE_SINE_SINE:
		sprintf((char*)buffer,
			"1 SINE  %#9.3fHz\n"
			"          %#6.2f%-4s\n"
			"2 SINE  %#9.3fHz\n"
			"          %#6.2f%-4s",
			gen_params[MODE2_FREQUENCY1].value, gen_params[MODE2_AMPLITUDE1].value, amp_unit,
			gen_params[MODE2_FREQUENCY2].value, gen_params[MODE2_AMPLITUDE2].value, amp_unit
		);
		break;

	case MODE_SWEEP:
		sprintf((char*)buffer,
			"1 SWEEP %#9.3fHz\n"
			"       \xc7%#9.3fHz\n"
			"%#5.3fs    %#6.2f%s\n"
			"%20s",
			gen_params[MODE3_FREQUENCY1].value, gen_params[MODE3_FREQUENCY2].value,
			gen_params[MODE3_TIME].value, gen_params[MODE3_AMPLITUDE].value, amp_unit,
			(gen_sweep_type == LIN) ? "LIN" : "LOG"
		);
		break;

	case MODE_BURST:
		sprintf((char*)buffer, "1 BURST  %#9.3fHz\n", gen_params[MODE4_FREQUENCY].value);
		if (gen_burst_mode == TIME) {
			sprintf((char*)(buffer + 21),
				"%#6.2f%-4s   %#5.3fs \n"
				"%#6.2f%-4s   %#5.3fs \n"
				"\xc7""cycles",
				gen_params[MODE4_AMPLITUDE1].value, amp_unit, gen_params[MODE4_TIME1].value,
				gen_params[MODE4_AMPLITUDE2].value, amp_unit, gen_params[MODE4_TIME2].value
			);
		}
		else {
			sprintf((char*)(buffer + 21),
				"%#6.2f%-4s  %5.0fcyc\n"
				"%#6.2f%-4s  %5.0fcyc\n"
				"\xc7seconds",
				gen_params[MODE4_AMPLITUDE1].value, amp_unit, gen_params[MODE4_CYCLES1].value,
				gen_params[MODE4_AMPLITUDE2].value, amp_unit, gen_params[MODE4_CYCLES2].value
			);
		}
		break;
	}

	display_write(buffer);
}

/****************************************************************************************************/

static void gen_change_param_amp_unit(uint32_t param, enum gen_t_amp_unit target_unit) {
	switch (target_unit) {
	case VRMS:
		gen_params[param].int_size	= 1;
		gen_params[param].min_value = 0.0f;
		gen_params[param].max_value	= DDS_MAX_AMPLITUDE;
		gen_params[param].value		= gen_scale_amplitude(gen_params[param].value);
		break;

	case VPEAK:
		gen_params[param].int_size	= 1;
		gen_params[param].min_value = 0.0f;
		gen_params[param].max_value	= M_SQRT2_F * DDS_MAX_AMPLITUDE;
		gen_params[param].value		= M_SQRT2_F * gen_scale_amplitude(gen_params[param].value);
		break;

	case DBFS:
		gen_params[param].int_size	= 2;
		gen_params[param].min_value = -99.99f;
		gen_params[param].max_value	= 0.0f;
		if (gen_scale_amplitude(gen_params[param].value) == 0) {
			gen_params[param].value = gen_params[param].min_value;
		} else {
			gen_params[param].value = 20.0f * log10f(
				gen_scale_amplitude(gen_params[param].value) / DDS_MAX_AMPLITUDE
				);
		}
		break;
	}
}

/****************************************************************************************************/

static void gen_change_amp_unit(enum gen_t_amp_unit target_amp_unit) {
	gen_change_param_amp_unit(MODE1_AMPLITUDE1, target_amp_unit);
	gen_change_param_amp_unit(MODE1_AMPLITUDE2, target_amp_unit);
	gen_change_param_amp_unit(MODE2_AMPLITUDE1, target_amp_unit);
	gen_change_param_amp_unit(MODE2_AMPLITUDE2, target_amp_unit);
	gen_change_param_amp_unit(MODE3_AMPLITUDE, target_amp_unit);
	gen_change_param_amp_unit(MODE4_AMPLITUDE1, target_amp_unit);
	gen_change_param_amp_unit(MODE4_AMPLITUDE2, target_amp_unit);

	gen_amp_unit = target_amp_unit;
}

/****************************************************************************************************/

static void gen_store_setting(uint8_t memory_slot) {
	struct gen_t_setting setting;
	setting.gen_mode		= gen_mode;
	setting.gen_amp_unit	= gen_amp_unit;
	setting.gen_sweep_type	= gen_sweep_type;
	setting.gen_burst_mode	= gen_burst_mode;

	for (uint32_t i = 0; i < GEN_PARAMS_COUNT; i++) {
		setting.gen_param_values[i] = gen_params[i].value;
	}

	memory_write(memory_slot * 128, (const uint8_t*)&setting, sizeof (setting));
}

/****************************************************************************************************/

static void gen_recall_setting(uint8_t memory_slot) {
	struct gen_t_setting setting;
	memory_read(memory_slot * 128, (uint8_t*)&setting, sizeof (setting));

	gen_mode		= setting.gen_mode;
	gen_sweep_type	= setting.gen_sweep_type;
	gen_burst_mode	= setting.gen_burst_mode;

	if (gen_amp_unit != setting.gen_amp_unit) {
		gen_change_amp_unit(setting.gen_amp_unit);
	}

	for (uint32_t i = 0; i < GEN_PARAMS_COUNT; i++) {
		gen_params[i].value = setting.gen_param_values[i];
	}

	gen_set_params();
	dds_set_mode((enum dds_t_mode)gen_mode);
	gen_update_display();
}

/****************************************************************************************************/

static void gen_cancel_input(void) {
	if (gen_input_mode == NUMERIC) {
		gen_input_param->value = gen_input_param_prev;	// restore previous value
	}
	display_hide_cursor();
	gen_input_param = NULL;	// exit input mode
}

/****************************************************************************************************/

static void gen_input(uint32_t param) {
	if (gen_input_param == &gen_params[param]) {
		// exit input mode for current parameter
		gen_cancel_input();
		return;
	}

	if (gen_input_param != NULL && gen_input_mode == NUMERIC) {
		// change currently changed parameter
		gen_input_param->value = gen_input_param_prev;	// restore previous value
	}

	gen_input_param = &gen_params[param];
	display_show_cursor(gen_input_param->display_pos);

	gen_input_mode = ENCODER;
	gen_cursor_exponent = 0;

	gen_input_part = INT;
	gen_input_pos = 0;
}

/****************************************************************************************************/

static uint32_t gen_check_input(void) {
	// check if value is in range
	if (gen_input_param->value < gen_input_param->min_value ||
		gen_input_param->value > gen_input_param->max_value) {
		return 0;
	}

	// perform additional check of parameters in burst mode
	if (gen_mode == MODE_BURST &&
		((
			gen_burst_mode == TIME &&
			(uint32_t)roundf((gen_params[MODE4_TIME1].value + gen_params[MODE4_TIME2].value) *
				DDS_SAMPLE_RATE > UINT32_MAX)
		)
		||
		(
			gen_burst_mode == CYCLES &&
			(gen_params[MODE4_CYCLES1].value + gen_params[MODE4_CYCLES2].value) *
				(uint32_t)roundf(DDS_SAMPLE_RATE / gen_params[MODE4_FREQUENCY].value) > UINT32_MAX
		))
		) {
		return 0;
	}

	return 1;
}

/****************************************************************************************************/

void gen_init(void) {
	dds_init();
	gen_set_params();
	gen_update_display();
}

/****************************************************************************************************/

void gen_process_button(uint32_t button) {
	float	previous_value;
	int8_t	digit = 0;

	switch (button) {
	case BUTTON_MODE:
		if (gen_function_mode != FUNCTION_GENERATOR) {
			break;
		}

		if (gen_input_param != NULL) {
			gen_cancel_input();
		}

		if (++gen_mode > MODE_BURST) {
			gen_mode = MODE_SINE_PHASE_SHIFT;
		}

		gen_set_params();
		dds_set_mode((enum dds_t_mode)gen_mode);
		break;

	case BUTTON_F1:
		if (gen_function_mode != FUNCTION_GENERATOR) {
			break;
		}

		switch (gen_mode) {
		case MODE_BURST: gen_input(MODE4_AMPLITUDE1); break;
		}
		break;

	case BUTTON_F2:
		if (gen_function_mode != FUNCTION_GENERATOR) {
			break;
		}

		switch (gen_mode) {
		case MODE_SWEEP:	gen_input(MODE3_TIME);			break;
		case MODE_BURST:	gen_input(MODE4_AMPLITUDE2);	break;
		}
		break;

	case BUTTON_F3:
		if (gen_function_mode != FUNCTION_GENERATOR) {
			break;
		}

		switch (gen_mode) {
		case MODE_BURST:
			if (gen_input_param != NULL) {
				gen_cancel_input();
			}

			gen_burst_mode = (gen_burst_mode == TIME) ? CYCLES : TIME;
			gen_set_params();
			break;
		}
		break;

	case BUTTON_F4:
		if (gen_function_mode != FUNCTION_GENERATOR) {
			break;
		}

		switch (gen_mode) {
		case MODE_SINE_PHASE_SHIFT:	gen_input(MODE1_FREQUENCY);		break;
		case MODE_SINE_SINE:		gen_input(MODE2_FREQUENCY1);	break;
		case MODE_SWEEP:			gen_input(MODE3_FREQUENCY1);	break;
		case MODE_BURST:			gen_input(MODE4_FREQUENCY);		break;
		}
		break;

	case BUTTON_F5:
		if (gen_function_mode != FUNCTION_GENERATOR) {
			break;
		}

		switch (gen_mode) {
		case MODE_SINE_PHASE_SHIFT:	gen_input(MODE1_AMPLITUDE1);	break;
		case MODE_SINE_SINE:		gen_input(MODE2_AMPLITUDE1);	break;
		case MODE_SWEEP:			gen_input(MODE3_FREQUENCY2);	break;
		case MODE_BURST:
			gen_input( (gen_burst_mode == TIME) ? MODE4_TIME1 : MODE4_CYCLES1 );
			break;
		}
		break;

	case BUTTON_F6:
		if (gen_function_mode != FUNCTION_GENERATOR) {
			break;
		}

		switch (gen_mode) {
		case MODE_SINE_PHASE_SHIFT:	gen_input(MODE1_PHASESHIFT);	break;
		case MODE_SINE_SINE:		gen_input(MODE2_FREQUENCY2);	break;
		case MODE_SWEEP:			gen_input(MODE3_AMPLITUDE);		break;
		case MODE_BURST:
			gen_input( (gen_burst_mode == TIME) ? MODE4_TIME2 : MODE4_CYCLES2 );
			break;
		}
		break;

	case BUTTON_F7:
		if (gen_function_mode != FUNCTION_GENERATOR) {
			break;
		}

		switch (gen_mode) {
		case MODE_SINE_PHASE_SHIFT:	gen_input(MODE1_AMPLITUDE2);	break;
		case MODE_SINE_SINE:		gen_input(MODE2_AMPLITUDE2);	break;
		case MODE_SWEEP:
			if (gen_input_param != NULL) {
				gen_cancel_input();
			}

			gen_sweep_type = (gen_sweep_type == LIN) ? LOG : LIN;	// change sweep type
			gen_set_params();
			break;
		}
		break;

	case BUTTON_AMP_UNIT:
		if (gen_function_mode != FUNCTION_GENERATOR) {
			break;
		}

		if (gen_input_param != NULL) {
			gen_cancel_input();
		}

		switch (gen_amp_unit) {
		case VRMS:	gen_change_amp_unit(VPEAK);	break;
		case VPEAK:	gen_change_amp_unit(DBFS);	break;
		case DBFS:	gen_change_amp_unit(VRMS);	break;
		}

		gen_set_params();
		break;

	case BUTTON_STORE_RECALL:
		if (gen_input_param != NULL) {
			gen_cancel_input();
		}

		if (++gen_function_mode > FUNCTION_SETTING_STORE) {
			gen_function_mode = FUNCTION_GENERATOR;
		}
		break;

	case BUTTON_OUTPUT_ENABLE_1:
		gen_output_enable1 ^= 1;
		dds_set_output_enable(1, gen_output_enable1);
		leds_toggle(0, gen_output_enable1);
		return;

	case BUTTON_OUTPUT_ENABLE_2:
		gen_output_enable2 ^= 1;
		dds_set_output_enable(2, gen_output_enable2);
		leds_toggle(1, gen_output_enable2);
		return;

	case BUTTON_DECIMAL_POINT:
		if (gen_input_param != NULL &&
			gen_input_part == INT && gen_input_param->fract_size > 0) {

			if (gen_input_mode == ENCODER) {
				// change input mode
				display_show_cursor(gen_input_param->display_pos);
				gen_input_param_prev = gen_input_param->value;	// store previous value
				gen_input_param->value = 0.0f;
				gen_input_mode = NUMERIC;
			}

			display_show_cursor(gen_input_param->display_pos + 2);	// move cursor to fractional part
			gen_input_pos = 0;
			gen_input_part = FRACT;
		}
		break;

	case BUTTON_ENTER:
		if (gen_input_param == NULL) {
			break;
		}

		if (!gen_check_input()) {
			gen_cancel_input();
		}

		gen_set_params();	// set new value

		display_hide_cursor();
		gen_input_param = NULL;	// exit input mode
		break;

	case BUTTON_ROTATE_LEFT:
		if (gen_input_param == NULL || gen_input_mode == NUMERIC) {
			return;	// to suppress display update
		}

		previous_value = gen_input_param->value;
		gen_input_param->value -= powf(10.0f, (float)gen_cursor_exponent);
		if (gen_check_input()) {
			gen_set_params();
		} else {
			gen_input_param->value = previous_value;
		}
		break;

	case BUTTON_ROTATE_RIGHT:
		if (gen_input_param == NULL || gen_input_mode == NUMERIC) {
			return;	// to suppress display update
		}

		previous_value = gen_input_param->value;
		gen_input_param->value += powf(10.0f, (float)gen_cursor_exponent);
		if (gen_check_input()) {
			gen_set_params();
		} else {
			gen_input_param->value = previous_value;
		}
		break;

	case BUTTON_LEFT:
		if (gen_input_param == NULL || gen_input_mode == NUMERIC) {
			break;
		}

		if (gen_cursor_exponent < gen_input_param->int_size - 1) {
			gen_cursor_exponent++;
		}
		if (gen_cursor_exponent >= 0) {
			display_show_cursor(gen_input_param->display_pos - gen_cursor_exponent);
		} else {
			display_show_cursor(gen_input_param->display_pos - gen_cursor_exponent + 1);
		}
		break;

	case BUTTON_RIGHT:
		if (gen_input_param == NULL || gen_input_mode == NUMERIC) {
			break;
		}

		if (gen_cursor_exponent > -gen_input_param->fract_size) {
			gen_cursor_exponent--;
		}
		if (gen_cursor_exponent >= 0) {
			display_show_cursor(gen_input_param->display_pos - gen_cursor_exponent);
		} else {
			display_show_cursor(gen_input_param->display_pos - gen_cursor_exponent + 1);
		}
		break;

	default:
		// numbers 0-9
		switch (button) {
		case BUTTON_1: digit = 1; break;
		case BUTTON_2: digit = 2; break;
		case BUTTON_3: digit = 3; break;
		case BUTTON_4: digit = 4; break;
		case BUTTON_5: digit = 5; break;
		case BUTTON_6: digit = 6; break;
		case BUTTON_7: digit = 7; break;
		case BUTTON_8: digit = 8; break;
		case BUTTON_9: digit = 9; break;
		}

		if (gen_input_param == NULL) {
			// handle number buttons in setting storage/recall mode
			switch (gen_function_mode) {
			case FUNCTION_SETTING_RECALL:
				gen_recall_setting(digit);
				gen_function_mode = FUNCTION_GENERATOR;
				break;
			case FUNCTION_SETTING_STORE:
				gen_store_setting(digit);
				gen_function_mode = FUNCTION_GENERATOR;
				break;
			}
			break;
		}

		if (gen_input_mode == ENCODER) {
			// change input mode
			display_show_cursor(gen_input_param->display_pos);
			gen_input_param_prev = gen_input_param->value;	// store previous value
			gen_input_param->value = 0.0f;
			gen_input_mode = NUMERIC;
		}

		if (gen_input_param->max_value <= 0.0f) {
			// if parameter should have negative value, change digit sign
			digit = -digit;
		}

		if (gen_input_part == INT) {
			if (gen_input_pos < gen_input_param->int_size) {
				if (gen_input_pos == 0 && digit == 0) {
					// leading zero has no effect
					break;
				}
				gen_input_param->value = 10 * gen_input_param->value + digit;
				gen_input_pos++;
			}
		} else {
			if (gen_input_pos < gen_input_param->fract_size) {
				gen_input_param->value += digit * powf(10.0f, (float)-(gen_input_pos + 1));
				if (++gen_input_pos != gen_input_param->fract_size) {
					// move cursor
					display_show_cursor(gen_input_param->display_pos + 2 + gen_input_pos);
				}
			}
		}
	}

	gen_update_display();
}

/****************************************************************************************************/
